import 'package:pdf/pdf.dart';
import 'package:pdf/widgets.dart';
import 'package:image/image.dart' as img;

import 'conversion.dart';
import 'generator_assets.dart';
import 'generator_card.dart';

/// Creates a PDF document from GeneratorCard lists.
class Generator {
  GeneratorAssets assets;

  /// Assets object links images used in the cards.
  Generator(this.assets);

  static const double mm = PdfPageFormat.mm;
  static const double pageHeight = 297 * mm;
  static const double pageWidth = 210 * mm;
  static const double cardHeight = 3.5 * PdfPageFormat.inch;
  static const double cardWidth = 2.5 * PdfPageFormat.inch;
  static const PdfPageFormat format = PdfPageFormat(pageWidth, pageHeight);
  Document _document;

  /// Initialises a document with the specified title.
  void openDocument(String title) {
    _document = Document(title: title);
  }

  /// Export the current document.
  List<int> saveDocument() {
    return _document.save();
  }

  /// Adds cards to the document. May create multiple pages of 9 cards each. Page order is not guaranteed.
  Future<void> addCards(List<GeneratorCard> cards) async {
    await Future.wait(List.generate((cards.length / 9).ceil(), (int i) {
      return cards.skip(i * 9).take(9).toList();
    }).map((List<GeneratorCard> pageCards) => addPage(pageCards)));
  }

  /// Adds a single page of up to 9 cards to the document. Throws an error if given more than 9 cards.
  Future<void> addPage(List<GeneratorCard> cards) async {
    _document ??
        (throw StateError(
            "Pages must not be added without opening a document with openDocument."));
    if (cards.length > 9) {
      throw ArgumentError("Pages must contain no more than 9 cards.");
    }

    List<Widget> cardWidgets = List.from(
        await Future.wait(
            cards.map((GeneratorCard card) => _buildCardWidget(card))),
        growable: true);

    while (cardWidgets.length < 9) {
      cardWidgets.add(
        SizedBox(
          height: cardHeight,
          width: cardWidth,
        ),
      );
    }
    _document.addPage(Page(
      pageFormat: format,
      build: (Context context) {
        return Center(
          child: Column(
            mainAxisAlignment: MainAxisAlignment.center,
            mainAxisSize: MainAxisSize.min,
            children: [
              Row(
                mainAxisAlignment: MainAxisAlignment.center,
                mainAxisSize: MainAxisSize.min,
                children: cardWidgets.take(3).toList(),
              ),
              Row(
                mainAxisAlignment: MainAxisAlignment.center,
                mainAxisSize: MainAxisSize.min,
                children: cardWidgets.skip(3).take(3).toList(),
              ),
              Row(
                mainAxisAlignment: MainAxisAlignment.center,
                mainAxisSize: MainAxisSize.min,
                children: cardWidgets.skip(6).take(3).toList(),
              ),
            ],
          ),
        );
      },
    ));
  }

  Future<Widget> _buildCardWidget(GeneratorCard card) async {
    return SizedBox(
      height: cardHeight,
      width: cardWidth,
      child: Stack(
        children: [
          Image(
            await _buildImage(assets.frameImages[card.element]),
          ),
          Stack(
            children: [
              Positioned(
                  left: 1.9 * mm,
                  top: 2.6 * mm,
                  child: Text(card.name, style: assets.title)),
              Positioned(
                  left: 56 * mm,
                  top: 1.5 * mm,
                  child: SizedBox(
                      height: 6 * mm,
                      width: 6 * mm,
                      child: Image(await _buildImage(
                          assets.elementImages[card.element])))),
              Positioned(
                  left: 1 * mm,
                  top: 8 * mm,
                  child: SizedBox(
                    height: 45 * mm,
                    width: 61.5 * mm,
                    child: Center(
                      child: Image(
                        await _buildImage(card.image),
                      ),
                    ),
                  )),
              Positioned(
                  left: 2 * mm,
                  top: 54.1 * mm,
                  child: Text(card.type, style: assets.type)),
              Positioned(
                right: 2 * mm,
                top: 54.1 * mm,
                child: Text(card.id, style: assets.id),
              ),
              Positioned(
                  left: 4.1 * mm,
                  top: 58.5 * mm,
                  child: SizedBox(
                      height: 24.1 * mm,
                      width: 55.4 * mm,
                      child: await _buildCardText(card.text))),
              Positioned(
                  left: 1.5 * mm,
                  bottom: 1.5 * mm,
                  child: Row(
                      mainAxisAlignment: MainAxisAlignment.start,
                      crossAxisAlignment: CrossAxisAlignment.end,
                      children: await _buildCostRow(card.cost))),
              Positioned(
                  right: 2 * mm,
                  bottom: 1.2 * mm,
                  child: Row(
                      mainAxisAlignment: MainAxisAlignment.end,
                      crossAxisAlignment: CrossAxisAlignment.end,
                      children: [Text(card.stats, style: assets.stats)]))
            ],
          ),
        ],
      ),
    );
  }

  Future<List<Widget>> _buildCostRow(Map<GeneratorElement, int> cost) async {
    Iterable<Future<List<Widget>>> a = cost.keys.map((key) async {
      return await Future.wait(
        List.generate(cost[key], (int i) async {
          return SizedBox(
            height: 4 * mm,
            width: 4 * mm,
            child: Image(
              await _buildImage(assets.elementImages[key]),
            ),
          );
        }),
      );
    });
    List<List<Widget>> b = await Future.wait(a);
    return b.expand((pair) => pair).toList();
  }

  Future<RichText> _buildCardText(String text) async {
    List<InlineSpan> result = List<InlineSpan>();

    int p = 0;

    for (int i = 0; i < text.length; i++) {
      if (text[i] == "<" || text[i] == "[") {
        if (i > p) {
          result.add(TextSpan(text: text.substring(p, i)));
        }
        p = i + 1;
      }
      if (text[i] == ">") {
        if (i > p) {
          result
              .add(TextSpan(text: text.substring(p, i), style: assets.keyword));
        }
        p = i + 1;
      }
      if (text[i] == "]") {
        if (i > p) {
          result.add(
            WidgetSpan(
              baseline: -0.3 * mm,
              child: Image(
                await _buildImage(
                  assets.elementImages[Conversion.elementFromString(
                      text.substring(p, i).toLowerCase())],
                ),
              ),
            ),
          );
        }
        p = i + 1;
      }
    }
    if (text.length > p) {
      result.add(TextSpan(text: text.substring(p, text.length)));
    }

    return RichText(text: TextSpan(children: result, style: assets.cardText));
  }

  Future<PdfImage> _buildImage(img.Image image) async {
    return PdfImage(_document.document,
        image: image.data.buffer.asUint8List(),
        width: image.width,
        height: image.height);
  }
}
